package cn.l7h.boing;

import javax.swing.*;
import java.awt.*;
import javax.sound.midi.*;

/**
 * Display a bouncing dot with sound inside a JComponent.
 * @author jl
 */
final class Dot {
	// frames per second
	private final static int FRAMERATE = 30;
	// diameter of the dot
	private final static int DOTSIZE = 10;
	private JComponent comp;
	private MidiDevice midiDev = null;

	private final class PosVel {
		double x = 0.0;
		double y = 60.0;
		double vx = 380.0; // pixels per second
		double vy = -250.0; // pixels per second
	}
	private PosVel myPosVel = new PosVel();

	private final class Clock extends Thread {
		private boolean toStop = false;
		
		void terminate() {
			toStop = true;
		}
		
		public void run() {
			while(!toStop) {
				recalc();						
				try {
					sleep(1000/FRAMERATE);
				}
				catch(InterruptedException e)
				{
					toStop = true;
				}
			}
		}
	}
	private Clock myClock = new Clock();

	Dot(JComponent c) {
		comp = c;

		try {
			midiDev = MidiSystem.getSynthesizer();
			midiDev.open();		
		}
		catch (MidiUnavailableException e)
		{}
	}

	void start() {
		myClock.start();		
	}

	void close() {
		myClock.terminate();
		midiDev.close();
	}

	void set(double x1, double y1, double dx1, double dy1) {
		synchronized(myPosVel) {
			myPosVel.x = x1;
			myPosVel.y = y1;
			myPosVel.vx = dx1;
			myPosVel.vy = dy1;
		}
	}

	private void recalc() {
		boolean boingX = false;
		boolean boingY = false;
		double vel = 0.0;
	
		synchronized(myPosVel) {
			// stop very slow motion
			if(Math.abs(myPosVel.y - comp.getHeight()) < 1.5
				&& Math.abs(myPosVel.vx) < 20.0 && Math.abs(myPosVel.vy) < 20.0) {
				return;
			}
			
			myPosVel.x += myPosVel.vx/FRAMERATE;			
			if(myPosVel.x<0.0) {
				myPosVel.x = 0.0;
				myPosVel.vx *= -1.0;
				boingX = true;
			} else if(myPosVel.x>comp.getWidth()-1.0) {
				myPosVel.x = comp.getWidth()-1.0;
				myPosVel.vx *= -1.0;
				boingX = true;
			}
					
			myPosVel.y += myPosVel.vy/FRAMERATE;
			if(myPosVel.y<0.0) {
				myPosVel.y = 0.0;
				myPosVel.vy *= -1.0;
				boingY = true;			
			} else if(myPosVel.y>comp.getHeight()-1.0) {
				myPosVel.y = comp.getHeight()-1.0;
				myPosVel.vy *= -1.0;
				boingY = true;			
			}
	
			myPosVel.vx *= 1.0-0.5/FRAMERATE;
			myPosVel.vy = (1.0-0.5/FRAMERATE)*myPosVel.vy + 1000.0/FRAMERATE;
			
			vel = 0.3*Math.sqrt(myPosVel.vx*myPosVel.vx+myPosVel.vy*myPosVel.vy);			
		}
		
		comp.repaint();
		
		if(midiDev != null && midiDev.isOpen() && (boingX || boingY)) {
			if (vel > 127.1) {
				vel = 127.1;
			} else if (vel < 1.1) {
				vel = 1.1;
			}
			ShortMessage note = new ShortMessage();
			try {
				note.setMessage(ShortMessage.NOTE_ON | 9, boingY ? 64 : 41, (int)vel);
			} catch (InvalidMidiDataException e)
			{}
			
			try {
				midiDev.getReceiver().send(note, -1);
			} catch (MidiUnavailableException e)
			{}	
		}
	}

	void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D)g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		Stroke s = new BasicStroke(DOTSIZE, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
		g2.setStroke(s);
		synchronized(myPosVel) {
			// How far do we move in one frame?
			double dist = Math.sqrt(myPosVel.vx*myPosVel.vx+myPosVel.vy*myPosVel.vy)/FRAMERATE; 
			// The faster, the lighter. And correct for gamma 2.2
			int grey = (int)(255.0*Math.pow(1.0 - DOTSIZE/(DOTSIZE+dist), 1.0/2.2));
			g2.setColor(new Color(grey, grey, grey));
			// draw half before and half behind the current position
			final double mix = 0.5/FRAMERATE;
			g2.drawLine((int)(myPosVel.x - mix*myPosVel.vx),
						(int)(myPosVel.y - mix*myPosVel.vy),
						(int)(myPosVel.x + mix*myPosVel.vx),
						(int)(myPosVel.y + mix*myPosVel.vy) );
		}
	}
}